Slime = {}
Slime.slimes = {} -- Holds common data for all slimes.
-- Make the list a global variable or it won't be saved properly
Slime_slimesList = {}

function Slime.run()

    -- Check if there is at least one slime eating the designated shape
    Slime.slimes.hunger.shapeStillThere = false

    Slime.slimes.physics.sticky.bodiesEaten = {}
    Slime.slimes.hunger.nextFoodSources = {}

    --> Process slimes.
    for i = 1, #Slime_slimesList do

        local slime = Slime_slimesList[i]

        -- Main stuff.
        Slime.slimes.physics.process(slime)
        Slime.slimes.hunger.process(slime)
        Slime.slimes.lifetime.process(slime)

        if regGetBool('perfo.freezeWhenEating') then
            local body = GetShapeBody(slime.sticky.shape)
            Slime.slimes.physics.sticky.bodiesEaten[body] = true
            Slime.slimes.physics.sticky.bodiesEatenOnce[body] = true
        end

    end

    -- Update food foodSources
    Slime.slimes.hunger.foodSources = Slime.slimes.hunger.nextFoodSources

    -- Unfreeze shapes no longer eaten by any slime
    if regGetBool('perfo.freezeWhenEating') then
        for k,v in pairs(Slime.slimes.physics.sticky.bodiesEatenOnce) do
            local body = k
            if not Slime.slimes.physics.sticky.bodiesEaten[body] then
                Slime.slimes.physics.sticky.bodiesEatenOnce[body] = nil
                if GetWorldBody() ~= body then
                    SetBodyDynamic(body, true)
                    SetBodyActive(body, true)
                end
            end
        end
    end

    -- If no slime eating this shape, reset shape to eat
    if not Slime.slimes.hunger.shapeStillThere then
        if Slime.slimes.hunger.shapeToEat == Slime.slimes.hunger.lastStaticShapeEaten then
            Slime.slimes.hunger.lastStaticShapeEaten = nil
        end
        Slime.slimes.hunger.shapeToEat = nil
    end

    -- Visual effects.
    Slime.slimes.effects.process()

    -- Auto-kill slimes to maintain FPS
    local preventDup = regGetBool('perfo.preventDup')
    local fps = FPS.value
    if fps < regGetFloat('perfo.killToMaintainFPS') and (not preventDup) then
        Slime.slimes.colony.markSlimesForRemoval()
    end

    -- Auto-kill slimes to respect max population
    local maxPop = regGetInt('perfo.slimesMaxPopulation')
    local slimePop = #Slime_slimesList
    if (maxPop > 0) and (slimePop > maxPop) and (not preventDup) then
        Slime.slimes.colony.markSlimesForRemoval(1 - (maxPop / slimePop))
    end

    -- Auto-kill Slimes attached to the same voxel
    if regGetBool('perfo.killStackedSlimes') then
        Slime.slimes.colony.removeStackedSlimes()
    end

    --> Safely delete specified slimes.
    Slime.slimes.colony.removeMarkedSlimes()

end


function Slime.slimes.getSize()
    if regGetBool('slime.scaleWithRadius') then
        return regGetFloat('slime.eatRadius')
    else
        return regGetFloat('slime.size')
    end
end



Slime.slimes.colony = {}

--- Create a slime object and store it in the list of slimes.
function Slime.slimes.colony.create(tr, parent)

    local generation = parent and (parent.generation + 1) or 1

    local slime = {

        id = #Slime_slimesList + 1,
        removeSlime = false, -- true = mark slime for removal/deletion.

        tr = tr, -- Position/rotation of the slime in the world
        foodEaten = 0, -- Food counter used to know when the slime should duplicate
        duplicationCount = 0, -- Number of duplications of this slime
        generation = generation, -- Generation of this slime
        uniqueShape = parent and parent.uniqueShape or nil, -- If slime eat only one shape, what shape to eat

        sticky = {
            shape = nil, -- Shape the slime is attached to or nil if falling
            shapeRelativePos = nil, -- Position of the slime relative to the shape position
            shapeVoxelIndex = nil, -- Index of the voxel the slime is attached to in shape coordinates
            shapeRelativeHitPos = nil, -- Position of the voxel the slime is attached to relative to the shape position
        },

        timers = {
            eatDelay =     {time = 9999, delay = regGetFloat('slime.eatDelay')}, -- Once slime is attached, wait before eating
            lifeDuration = {time = 9999, delay = regGetFloat('slime.lifeDuration')}, -- Life duration fo the slime (if not 0)
        }

    }

    -- Eat timer varience.
    local eatDelay = regGetFloat('slime.eatDelay')
    slime.timers.eatDelay.delay = rdmf(eatDelay * 0.9, eatDelay * 1.1) -- Registry live updates.

    -- Initialize time remaining.
    TimerResetTime(slime.timers.eatDelay)
    TimerResetTime(slime.timers.lifeDuration)

    table.insert(Slime_slimesList, slime)
    dbp('Created slime. ' .. sfnTime()..', eatDelay='..eatDelay..', rndDelay='..slime.timers.eatDelay.delay)

end

--- Spawn a slime into the world.
function Slime.slimes.colony.spawn(pos, parent)

    local generation = parent and (parent.generation + 1) or 1
    Slime.slimes.colony.create(Transform(pos, QuatEuler(0,0,0)), parent)

    if generation <= 1 then
        if rdm() < 0.1 / regGetFloat('tool.pour.rate') and regGetBool('slime.soundOn') then
            sounds.play.slime(pos, 0.5)
        end

        SpawnParticle("smoke", pos, Vec(0,0,0), 0.5, 0.5, 0.5, 0.5)
    end

end

--- Mark a slime for removal to safely remove it near the end of the frame.
function Slime.slimes.colony.markSlimeForRemoval(slime)
    slime.removeSlime = true
    dbp('Slime marked for removal: ' .. slime.id .. ' ' ..sfnTime())
end

--- Mark half of the slimes for removal.
function Slime.slimes.colony.markSlimesForRemoval(percent)
    percent = percent or 0.5
    local nbSlimes = #Slime_slimesList
    local nbSlimesToDelete = math.floor(nbSlimes * percent)
    if nbSlimesToDelete > 0 then
        for i = 1, nbSlimesToDelete do
            Slime.slimes.colony.markSlimeForRemoval(Slime_slimesList[i])
        end
    end
end

--- Mark slimes attached to the same voxel for removal.
function Slime.slimes.colony.removeStackedSlimes()
    local attachedSlimesSet = {}
    for i = #Slime_slimesList, 1, -1 do
        local slime = Slime_slimesList[i]
        if slime.sticky.shape ~= nil then
            local v = slime.sticky.shapeVoxelIndex
            local attachmentId = "" .. slime.sticky.shape .."-"..v[1]..","..v[2]..","..v[3]
            if attachedSlimesSet[attachmentId] == nil then
                attachedSlimesSet[attachmentId] = true
            else
                Slime.slimes.colony.markSlimeForRemoval(slime)
            end
        end
    end
end


--- Safely delete all slimes that are marked for removal.
function Slime.slimes.colony.removeMarkedSlimes()

    TableRemove(Slime_slimesList, function(t, i, j)
        -- Return true to keep the value, or false to discard it.
        return not Slime_slimesList[i].removeSlime;
    end)
    
end

--- Get a valid spawn or jump position for a slime
function Slime.slimes.colony.getSpawnPos(center, radius)

    local spawnPos = getRandPointOnSphere(center, radius)
    -- Try to avoid spawning slime inside unbreakable materials/shapes as they would die instantly
    local hit, point, normal, shape = QueryClosestPoint(spawnPos, 0.1)
    if hit then
        local material = GetShapeMaterialAtPosition(shape, spawnPos)
        -- dbp('SLIME would spawn in material '..material)
        if IsMaterialUnbreakable(material, shape) then
            -- dbp('SLIME spawn point moved')
            -- Material, shape or body is unbreakable, move spawnPos to the opposite point on the sphere instead
            spawnPos = VecAdd(spawnPos, VecScale(VecSub(center, spawnPos), 2))
        end
    end

    return spawnPos

end





Slime.slimes.physics = {}

---Process the physics of a single slime.
function Slime.slimes.physics.process(slime)

    if slime.sticky.shape == nil then -- Right after gas projectile hits something.

        Slime.slimes.physics.sticky.dripAndStick(slime) -- Let the slime drip and find a new shape

    elseif Slime.slimes.physics.sticky.didStickPointBreak(slime) then -- Slime shape broke previous frame.

        -- local v = slime.sticky.shapeVoxelIndex
        -- local hitPos = TransformToParentPoint(GetShapeWorldTransform(slime.sticky.shape), slime.sticky.shapeRelativeHitPos)
        -- dbp('Slime ' .. slime.id..' mass decreased: stick point='..VecStr(slime.sticky.shapeVoxelIndex)..', world pos='..VecStr(hitPos)..', material='..GetShapeMaterialAtIndex(slime.sticky.shape, v[1], v[2], v[3]))
        -- dbl(hitPos, VecAdd(hitPos, Vec(0, 0.01, 0)), 1, 1, 1, 1)
        Slime.slimes.physics.sticky.dripAndStick(slime) -- QCP right after shape breaks.

    else -- Slime is stuck to a shape and did not break last frame.

        -- Place slime pos relative to shape sticky point.
        slime.tr.pos = TransformToParentPoint(GetShapeWorldTransform(slime.sticky.shape), slime.sticky.shapeRelativePos)
        -- local v = slime.sticky.shapeVoxelIndex
        -- local hitPos = TransformToParentPoint(GetShapeWorldTransform(slime.sticky.shape), slime.sticky.shapeRelativeHitPos)
        -- dbp('Slime ' .. slime.id..' stick point='..VecStr(slime.sticky.shapeVoxelIndex)..', world pos='..VecStr(hitPos)..', material='..GetShapeMaterialAtIndex(slime.sticky.shape, v[1], v[2], v[3]))
        -- dbl(hitPos, VecAdd(hitPos, Vec(0, 0.01, 0)), 1, 1, 1, 1)

    end

    -- Debug.
    dbw('SLIME '.. slime.id ..' attached to voxel ', slime.sticky.shapeVoxelIndex and VecStr(slime.sticky.shapeVoxelIndex) or "nil")

end


Slime.slimes.physics.sticky = {}
Slime.slimes.physics.sticky.bodiesEaten = {}
Slime.slimes.physics.sticky.bodiesEatenOnce = {}

function Slime.slimes.physics.sticky.setShape(slime, shape, hitPoint)

    slime.sticky.shape = shape
    slime.sticky.shapeRelativePos = TransformToLocalPoint(GetShapeWorldTransform(shape), slime.tr.pos)
    -- slime.sticky.shapeVoxelCount = GetShapeVoxelCount(slime.sticky.shape)
    slime.sticky.shapeVoxelIndex = GetVoxelIndex(shape, hitPoint)
    local xsize, ysize, zsize, scale = GetShapeSize(shape)
    slime.sticky.shapeVoxelBox = Vec(xsize, ysize, zsize)
    slime.sticky.shapeRelativeHitPos = TransformToLocalPoint(GetShapeWorldTransform(shape), hitPoint)
    -- local v = slime.sticky.shapeVoxelIndex
    -- dbp('Slime ' .. slime.id..' strick: stick point='..VecStr(slime.sticky.shapeVoxelIndex)..', hitPoint='..VecStr(hitPoint)..', material='..GetShapeMaterialAtIndex(slime.sticky.shape, v[1], v[2], v[3]))
    -- dbl(hitPoint, VecAdd(hitPoint, Vec(0, 0.01, 0)), 1, 1, 1, 1)

    if regGetBool('perfo.eatOnlyFirst') then
        if slime.uniqueShape == nil then
            slime.uniqueShape = shape
        elseif shape ~= slime.uniqueShape then
            Slime.slimes.colony.markSlimeForRemoval(slime)
        end
    end

    if regGetBool('perfo.freezeWhenEating') then
        local body = GetShapeBody(slime.sticky.shape)
        Slime.slimes.physics.sticky.bodiesEaten[body] = true
        Slime.slimes.physics.sticky.bodiesEatenOnce[body] = true
        if GetWorldBody() ~= body then
            SetBodyDynamic(body, false)
        end
    end
    
    -- dbp('slime shape set ' .. sfnTime())
end

function Slime.slimes.physics.sticky.resetShape(slime)
    -- if IsHandleValid(slime.sticky.shape) then
    --     local v = slime.sticky.shapeVoxelIndex
    --     local hitPos = TransformToParentPoint(GetShapeWorldTransform(slime.sticky.shape), slime.sticky.shapeRelativeHitPos)
    --     dbp('Slime ' .. slime.id..' shape destroyed: stick point='..VecStr(slime.sticky.shapeVoxelIndex)..', world pos='..VecStr(hitPos)..', material='..GetShapeMaterialAtIndex(slime.sticky.shape, v[1], v[2], v[3]))
    --     dbl(hitPos, VecAdd(hitPos, Vec(0, 0.01, 0)), 1, 1, 1, 1)
    -- end
    slime.sticky.shape = nil -- Set no shape
    slime.sticky.shapeRelativePos = nil
    slime.sticky.shapeVoxelIndex = nil
    -- slime.sticky.shapeVoxelCount = nil
    slime.sticky.shapeVoxelBox = nil
    -- dbp('slime shape reset ' .. sfnTime())
end

function Slime.slimes.physics.sticky.didStickPointBreak(slime)
    -- return GetShapeVoxelCount(slime.sticky.shape) ~= slime.sticky.shapeVoxelCount
    
    local xsize, ysize, zsize, scale = GetShapeSize(slime.sticky.shape)
    local vb = slime.sticky.shapeVoxelBox
    local shapeUnchanged = (vb[1] == xsize) and (vb[2] == ysize) and (vb[3] == zsize)
    if not shapeUnchanged then
        return true
    end
    local v = slime.sticky.shapeVoxelIndex
    return GetShapeMaterialAtIndex(slime.sticky.shape, v[1], v[2], v[3]) == ''
end

--- Check the surroundings of a slime and attach it to very close objects.
function Slime.slimes.physics.sticky.dripAndStick(slime)

    --- If single shape to eat and shape is no longer there, kill slime.
    if regGetBool('perfo.eatOnlyFirst') and slime.uniqueShape then
        if not IsShapeValid(slime.uniqueShape) then
            Slime.slimes.colony.markSlimeForRemoval(slime)
            return
        end
    end

    local slimeRcHit, point, normal, shape = QueryClosestPoint(slime.tr.pos, Slime.slimes.getSize())
    if slimeRcHit then
        -- local material = GetShapeMaterialAtPosition(slimeRcShape, point)
        -- dbp('SLIME '..slime.id..' stick to material '..material)
        -- Set shape.
        Slime.slimes.physics.sticky.setShape(slime, shape, point)

    else --> No hit.

        local hit, point, normal, shape = QueryClosestPoint(slime.tr.pos, regGetFloat('slime.jumpDistance'))
        if hit then
            local targetVec = VecSub(point, slime.tr.pos)
            local dist = math.min(VecLength(targetVec), 0.5)
            slime.tr.pos = VecAdd(slime.tr.pos, VecScale(VecNormalize(targetVec), dist))
        else
            -- If none found, drip down.
            slime.tr.pos = VecAdd(slime.tr.pos, Vec(0, -0.5, 0))
        end
        
        -- No shape.
        Slime.slimes.physics.sticky.resetShape(slime)
        -- Slime moved, reset eat timer
        TimerResetTime(slime.timers.eatDelay)

    end

end

function Slime.slimes.physics.sticky.jumpTo(slime, pos)

    -- Only jump if landed
    if slime.sticky.shape == nil then
        return
    end
    -- Enter free fall mode
    Slime.slimes.physics.sticky.resetShape(slime)
    -- Jump 45deg up in the direction of pos
    local jumpDir = Vec(1, 1, 0)
    local currentFlatDir = VecFlatten(jumpDir)
    local expectedFlatDir = VecFlatten(VecSub(pos, slime.tr.pos))
    local horizontalAngle = Vec2DOrientedAngleBetween(Vec3DTo2D(currentFlatDir, "y"), Vec3DTo2D(expectedFlatDir, "y"))
    jumpDir = VecRotate(jumpDir, "y", horizontalAngle)
    -- Jump a little further than max jump to avoid slime to immediately reattach to the same voxel on fall
    jumpDir = VecResize(jumpDir, regGetFloat('slime.jumpDistance') * 2)
    slime.tr.pos = VecAdd(slime.tr.pos, jumpDir)

end




Slime.slimes.hunger = {}

Slime.slimes.hunger.shapeToEat = nil
Slime.slimes.hunger.lastStaticShapeEaten = nil
Slime.slimes.hunger.shapeStillThere = false
Slime.slimes.hunger.foodSources = {}
Slime.slimes.hunger.nextFoodSources = {}

--- Processes the hunger of a slime.
function Slime.slimes.hunger.process(slime)

    Slime.slimes.hunger.eat(slime)

end

-- Eat when slime is sticked to something and eat timer is over.
function Slime.slimes.hunger.eat(slime)
    -- Slime is sticked to something and is allowed to eat it
    if slime.sticky.shape ~= nil then
        -- If only one shape shold be eaten at a time
        if regGetBool('perfo.eatOneAtATime') then
            -- Set the shape to eat if none was already there
            if Slime.slimes.hunger.shapeToEat == nil then
                if Slime.slimes.hunger.lastStaticShapeEaten then
                    Slime.slimes.hunger.shapeToEat = Slime.slimes.hunger.lastStaticShapeEaten
                else
                    Slime.slimes.hunger.shapeToEat = slime.sticky.shape
                    if not IsBodyDynamic(GetShapeBody(Slime.slimes.hunger.shapeToEat)) then
                        Slime.slimes.hunger.lastStaticShapeEaten = Slime.slimes.hunger.shapeToEat
                    end
                end
            else
                -- Do not eat if sticking to another shape
                if slime.sticky.shape ~= Slime.slimes.hunger.shapeToEat then
                    -- If eating a static shape, give preference to eat dynamic shape instead
                    if not IsBodyDynamic(GetShapeBody(Slime.slimes.hunger.shapeToEat)) and
                    IsBodyDynamic(GetShapeBody(slime.sticky.shape)) then
                        Slime.slimes.hunger.shapeToEat = slime.sticky.shape
                    else
                        return
                    end
                end
            end
        end
        -- If only eating when shape is not moving
        if regGetBool('perfo.noEatingWhileMoving') then
            local body = GetShapeBody(slime.sticky.shape)
            if VecLength(GetBodyVelocity(body)) > 0.1 then
                return
            end
        end

        Slime.slimes.hunger.shapeStillThere = true

        TimerRunTime(slime.timers.eatDelay)

        if slime.timers.eatDelay.time <= 0 then
            -- Eat matter nearby
            local eatSuccess = false
            local size = Slime.slimes.getSize()
            -- Found something to eat, try to eat it (may be indestructible)
            local point = TransformToParentPoint(GetShapeWorldTransform(slime.sticky.shape), slime.sticky.shapeRelativeHitPos)
            dbp('SLIME '.. slime.id ..' try eating at ' .. VecStr(point))
            local eatRadius = regGetFloat('slime.eatRadius')
            local voxelsDeleted = MakeHole(point, eatRadius, eatRadius, eatRadius, not regGetBool('slime.soundOn'))
            -- If eat was successful (voxels were deleted by MakeHole)
            if voxelsDeleted > 0 then
                eatSuccess = true
                slime.foodEaten = slime.foodEaten + 1
                -- Duplicate slime if current slime ate enough food
                local duplicationRate = regGetInt('slime.duplicationRate')
                -- And if max duplication is not reached (max 100 = infinite)
                local maxDuplication = regGetInt('perfo.slimesMaxDuplication')
                -- And Slime is not of max generation
                local maxGeneration = regGetInt('perfo.slimesMaxGeneration')
                dbp('SLIME '.. slime.id ..' eating success, food='..slime.foodEaten..'/'..duplicationRate..', duplications='..slime.duplicationCount..'/'..maxDuplication..', generation='..slime.generation..'/'..maxGeneration)
                if slime.foodEaten >= duplicationRate
                and (maxDuplication == 100 or slime.duplicationCount < maxDuplication) 
                and (maxGeneration == 0 or slime.generation < maxGeneration)
                and (not Slime.slimes.hunger.shouldSkipDuplication()) then
                    slime.foodEaten = 0
                    local spawnPos = Slime.slimes.colony.getSpawnPos(point, eatRadius)
                    dbp('SLIME '.. slime.id ..' duplicate center='..VecStr(point)..', spawnPos='..VecStr(spawnPos))
                    Slime.slimes.colony.spawn(spawnPos, slime)
                    -- Kill slime if max duplication reached and this option is enabled
                    slime.duplicationCount = slime.duplicationCount + 1
                    if regGetBool('perfo.killSlimeAfterLastDup') and (maxDuplication ~= 100) then
                        if slime.duplicationCount >= maxDuplication then
                            Slime.slimes.colony.markSlimeForRemoval(slime)
                        end
                    end

                end
                -- Detach from body / reset eat timer
                TimerResetTime(slime.timers.eatDelay)
                dbp('SLIME '.. slime.id ..' timer reset to '..slime.timers.eatDelay.time..', delay='..slime.timers.eatDelay.delay)
                Slime.slimes.physics.sticky.resetShape(slime)
                -- If hive mind, remember food source for next loop
                if regGetBool('slime.hiveMind') then
                    Slime.slimes.hunger.nextFoodSources[#Slime.slimes.hunger.nextFoodSources+1] = point
                end
                -- If eating player
                if regGetBool('slime.eatPlayer') then
                    local playerHead = GetPlayerCameraTransform().pos
                    local playerFeet = GetPlayerTransform().pos
                    local eatSpotToPlayer = VecSub(playerFeet, point)
                    local flatDistToPlayer = VecLength(VecFlatten(eatSpotToPlayer))
                    if flatDistToPlayer <= eatRadius and point[2] <= (playerHead[2] + eatRadius) and point[2] >= (playerFeet[2] - eatRadius) then
                        SetPlayerHealth(GetPlayerHealth() - 0.1)
                    end
                end
            end
            -- If eat failed, slime dies
            if (not eatSuccess) and (not Slime.slimes.hunger.tryJumpToFoodSource(slime)) and (not regGetBool('perfo.gooMode')) then
                -- If eat failed, slime dies
                Slime.slimes.colony.markSlimeForRemoval(slime)
            end

        end

    end

end

--- If hive mind is active, try to jump in the direction of a known food source
function Slime.slimes.hunger.tryJumpToFoodSource(slime)
    local nbFoodSources = #Slime.slimes.hunger.foodSources
    if (not regGetBool('slime.hiveMind')) or nbFoodSources == 0 then
        return false
    end

    local targetFoodSource = Slime.slimes.hunger.foodSources[rdm(1, nbFoodSources)]
    Slime.slimes.physics.sticky.jumpTo(slime, targetFoodSource)
    return true
end

--- Returns if the colony is overpopulated
function Slime.slimes.hunger.shouldSkipDuplication()
    -- Skip if goo mode is active
    if regGetBool('perfo.gooMode') then
        return true
    end

    -- Never skip if prevent duplication is false
    if not regGetBool('perfo.preventDup') then
        return false
    end

    -- Check if FPS limit reached
    local fps = FPS.value
    if fps < regGetFloat('perfo.killToMaintainFPS') then
        return true
    end

    -- Check if max population reached
    local maxPop = regGetInt('perfo.slimesMaxPopulation')
    local slimePop = #Slime_slimesList
    if (maxPop > 0) and (slimePop > maxPop) then
        return true
    end

    -- Else duplication can happen
    return false
end






Slime.slimes.lifetime = {}

function Slime.slimes.lifetime.process(slime)
    -- Life duration of 0 is unlimited
    if regGetFloat('slime.lifeDuration') > 0 then
        TimerRunTime(slime.timers.lifeDuration)
        -- If life duration runs out, kill the slime
        if slime.timers.lifeDuration.time <= 0 then
            Slime.slimes.colony.markSlimeForRemoval(slime)
        end
    end
end





Slime.slimes.effects = {}

function Slime.slimes.effects.process()

    -- Render slimes.
    for i = 1, #Slime_slimesList do

        local slime = Slime_slimesList[i]
        
        if regGetBool('perfo.renderSlimeParticles') then
            Slime.slimes.effects.renderSlimeIdle(slime.tr.pos) -- Idle slime
        end

    end


    -- Render slime projectiles.
    for i = 1, #projectiles do

        local proj = projectiles[i]
        local r = regGetFloat('slime.color.r')/255
        local g = regGetFloat('slime.color.g')/255
        local b = regGetFloat('slime.color.b')/255
        local a = regGetFloat('slime.color.a')/255
        local size = Slime.slimes.getSize()
        DrawDot(proj.transform.pos, size, size, r, g, b, a) -- Slime projectile

    end

end

--- Draw ui dots instead of particles based on.
function Slime.slimes.effects.renderSlimesIdleSimple()

    do UiPush()

        local camTr = GetCameraTransform()

        for i = 1, #Slime_slimesList do

            local slime = Slime_slimesList[i]
            local dist = VecDist(camTr.pos, slime.tr.pos)
            local size = Slime.slimes.getSize()
            local s = (90/dist) * size * 10

            -- Check if slime is in front of the camera.
            if TransformToLocalPoint(camTr, slime.tr.pos)[3] < 0 then

                local slimePx, slimePy = UiWorldToPixel(slime.tr.pos)

                do UiPush()
                    
                    local r = regGetFloat('slime.color.r')/255
                    local g = regGetFloat('slime.color.g')/255
                    local b = regGetFloat('slime.color.b')/255
                    local a = regGetFloat('slime.color.a')/255
                    UiColor(r, g, b, a)
                    UiTranslate(slimePx - (s/2), slimePy - (s/2))
                    -- UiImage('ui/common/dot.png')
                    UiRect(s,s)

                UiPop() end

            end

        end

    UiPop() end

end

function Slime.slimes.effects.renderSlimeIdle(pos)

    ParticleReset()

    -- Idle slime particles.
    local r = regGetFloat('slime.color.r')/255
    local g = regGetFloat('slime.color.g')/255
    local b = regGetFloat('slime.color.b')/255
    local a = regGetFloat('slime.color.a')/255
    local glow = regGetFloat('slime.glow')
    local size = Slime.slimes.getSize()
    local particleType = "plain"
    if regGetBool('perfo.particleTypeSmoke') then
        particleType = "smoke"
    end
    ParticleType(particleType)
    ParticleTile(4)
    ParticleColor(r, g, b)
    ParticleRadius(size/2.5)
    ParticleAlpha(a)
    ParticleGravity(0)
    ParticleDrag(0)
    ParticleEmissive(glow)
    ParticleRotation(0)
    ParticleStretch(0)
    ParticleSticky(0)
    ParticleCollide(0)

    SpawnParticle(pos, Vec(), 0.05)

end